/*
** $Id: vector.c,v 0.1 19.10.2017 $
** vector dynamic array Library
** See Copyright Notice in agena.h
*/

/* Just demonstrates the usage of the vector.c dynamic C array library */
/* You CANNOT pass DOUBLES or other simple 8-byte data types via (void *) arguments, since pointers are usually 4 bytes wide.
   Put them into unions or arrays, alternatively. Casting doubles to pointers like "double *p = (double *)num" and back
   to "double num = *p;" is not portable.

   See also: https://stackoverflow.com/questions/6575340/converting-double-to-void-in-c
             https://stackoverflow.com/questions/12700384/convert-void-to-double */

#define dyvec_c
#define LUA_LIB

#include <stdlib.h>
#include <string.h>

#include "vector.h"
#include "agena.h"
#include "agnxlib.h"

#define checkarray(L, n) (vector *)luaL_checkudata(L, n, "dyvec")

#if !(defined(LUA_DOS) || defined(__OS2__) || defined(LUA_ANSI))
#define AGENA_DYVECLIBNAME "dyvec"
LUALIB_API int (luaopen_dyvec) (lua_State *L);
#endif


static int dyvec_int32 (lua_State *L) {
  vector *a;
  a = (vector *)lua_newuserdata(L, sizeof(vector));
  if (sizeof(int32_t) != sizeof(void *) || sizeof(int) != sizeof(void *))
    luaL_error(L, "Error in " LUA_QS ": int32_t's not supported on this platform.", "dyvec.int32");
  if (vector_init(a, sizeof(int32_t)))
    luaL_error(L, "Error in " LUA_QS ": memory allocation error.", "dyvec.int32");
  luaL_getmetatable(L, "dyvec");
  lua_setmetatable(L, -2);
  lua_pushstring(L, AGENA_DYVECLIBNAME);
  agn_setutype(L, -2, -1);
  agn_poptop(L);
  return 1;
}


static int dyvec_append (lua_State *L) {  /* append a number to the tail. */
  vector *a = checkarray(L, 1);
  int32_t x = agn_checkinteger(L, 2);
  if (vector_append(a, (void *)x) == 1)
    luaL_error(L, "Error in " LUA_QS ": memory allocation error.", "dyvec.append");
  return 0;
}


/* dyvec.setitem (a, i, v)
   With a any dyvec, sets number v to a[i], where i, the index, is an integer counting from 1. */
static int dyvec_setitem (lua_State *L) {
  vector *a = checkarray(L, 1);
  int64_t index = agn_checkinteger(L, 2) - 1;  /* realign Lua/Agena index to C index, starting from 0 */
  int32_t x = agn_checkinteger(L, 3);
  if (index < 0 || index >= a->size)
    luaL_error(L, "Error in " LUA_QS ": index %d out of range.", "dyvec.put", index + 1);
  vector_set(a, index, (void *)x);
  return 0;
}


/* dyvec.put (a, i, v)
   With a any dyvec, inserts number v to a[i], shifting all elements to open space. i, the index, is an 
   integer counting from 1. A value can also be appended with this function. */
static int dyvec_put (lua_State *L) {
  vector *a = checkarray(L, 1);
  int64_t index = agn_checkinteger(L, 2) - 1;  /* realign Lua/Agena index to C index, starting from 0 */
  int32_t x = agn_checkinteger(L, 3);
  if (index < 0 || index > a->size)
    luaL_error(L, "Error in " LUA_QS ": index %d out of range.", "dyvec.put", index + 1);
  if (vector_insert(a, index, (void *)x) == 1)
    luaL_error(L, "Error in " LUA_QS ": memory allocation error.", "dyvec.put");
  return 0;
}


static int dyvec_purge (lua_State *L) {
  vector *a = checkarray(L, 1);
  int64_t index = agn_checkinteger(L, 2) - 1;  /* realign Lua/Agena index to C index, starting from 0 */
  if (index < 0 || index >= a->size)
    luaL_error(L, "Error in " LUA_QS ": index %d out of range.", "dyvec.purge", index + 1);
  if (vector_delete(a, index) == 1)
    luaL_error(L, "Error in " LUA_QS ": memory allocation error.", "dyvec.purge");
  return 0;
}


/* dyvec.get (a, i)
  With a any dyvec, returns the value stored at a[i], where i, the index, is an integer counting from 1. */
static int dyvec_getitem (lua_State *L) {
  vector *a = checkarray(L, 1);
  int64_t index = agn_checkinteger(L, 2) - 1;  /* realign Lua/Agena index to C index, starting from 0 */
  if (index < 0 || index >= a->size)
    luaL_error(L, "Error in " LUA_QS ": index %d out of range.", "dyvec.get", index + 1);
  lua_pushinteger(L, (int32_t)vector_get(a, index));
  return 1;
}


/* dyvec.iterate (a [, i [, p]])
Returns an iterator function that when called returns the next value in the dyvec userdata structure
a, or null if there are no further entries in the structure.

If an index i is passed, the first call to the iterator function returns the i-th element in
the dyvec list and with subsequent calls, the respective elements after index i.

You may also pass a positive integer step p to the iterator function: If given, then in subsequent
calls the p-th element after the respective current one is returned, equivalent to giving an optional
step size in numeric for loops.

Example:

> import dyvec

> a := dyvec.int32(3)

> for i to 3 do a[i] := 2*i od

> f := dyvec.iterate(a, 2):  # return all values starting with index 2
procedure(01CDC200)

> f():
1

> f():
2

> f():
3

> f():  # no more values in a
null */

static int iterate (lua_State *L) {
  vector *a;
  size_t pos, step;
  a = lua_touserdata(L, lua_upvalueindex(1));
  pos = lua_tointeger(L, lua_upvalueindex(2));
  step = lua_tointeger(L, lua_upvalueindex(3));
  if (pos > a->size)
    lua_pushnil(L);
  else {
    lua_pushinteger(L, (int32_t)vector_get(a, pos - 1));
    lua_pushinteger(L, pos + step);
    lua_replace(L, lua_upvalueindex(2));
  }
  return 1;
}

static int dyvec_iterate (lua_State *L) {
  int64_t pos, step;
  vector *a = checkarray(L, 1);
  pos = agnL_optinteger(L, 2, 1);
  step = agnL_optinteger(L, 3, 1);
  if (pos < 1 || (pos > a->size && a->size != 0))  /* when iterating, return null if the array is empty */
    luaL_error(L, "Error in " LUA_QS ": index %d out of range.", "dyvec.iterate", pos);
  if (step < 1)
    luaL_error(L, "Error in " LUA_QS ": step size must be a positive.", "dyvec.iterate");
  lua_pushvalue(L, 1);
  lua_pushinteger(L, pos);
  lua_pushinteger(L, step);
  lua_pushcclosure(L, &iterate, 3);
  return 1;
}


static int mt_getsize (lua_State *L) {  /* returns the number of allocated slots in the userdata numeric array */
  vector *a = checkarray(L, 1);
  lua_pushinteger(L, a->size);
  return 1;
}


static int mt_dyvec2string (lua_State *L) {  /* at the console, the array is formatted as follows: */
  vector *a;
  a = checkarray(L, 1);
  if (agn_getutype(L, 1)) {
    lua_pushfstring(L, "(%d)", a->size);
    lua_concat(L, 2);
  } else
    luaL_error(L, "Error in " LUA_QS ": invalid type.", "dyvec.__tostring");
  return 1;
}


static int mt_dyvecgc (lua_State *L) {  /* please do not forget to garbage collect deleted userdata */
  (void)L;
  vector *a = checkarray(L, 1);
  vector_free(a);
  return 0;
}


/* dyvec.toseq (a)
   Receives a dyvec userdata structure a and converts it into a sequence of numbers, the return. */
static int dyvec_toseq (lua_State *L) {
  size_t i;
  vector *a = checkarray(L, 1);
  agn_createseq(L, a->size);
  for (i=0; i < a->size; i++)
    agn_seqsetinumber(L, -1, i + 1, (int32_t)vector_get(a, i));
  return 1;
}


static const struct luaL_Reg dyvec_arraylib [] = {  /* metamethods for numeric userdata `n` */
  {"__index", dyvec_getitem},       /* n[p], with p the index, counting from 1 */
  {"__writeindex", dyvec_setitem},  /* n[p] := value, with p the index, counting from 1 */
  {"__tostring", mt_dyvec2string},  /* for output at the console, e.g. print(n) */
  {"__size", mt_getsize},
  {"__gc", mt_dyvecgc},             /* please do not forget garbage collection */
  {NULL, NULL}
};


static const luaL_Reg dyveclib[] = {
  {"append",  dyvec_append},
  {"int32",   dyvec_int32},
  {"getitem", dyvec_getitem},
  {"iterate", dyvec_iterate},
  {"purge",   dyvec_purge},
  {"put",     dyvec_put},
  {"setitem", dyvec_setitem},
  {"toseq",   dyvec_toseq},
  {NULL, NULL}
};


/*
** Open dyvec library
*/
LUALIB_API int luaopen_dyvec (lua_State *L) {
  luaL_newmetatable(L, "dyvec");  /* metatable for dyvec userdata, adds it to the registry with key 'dyvec' */
  luaL_register(L, NULL, dyvec_arraylib);  /* assign C metamethods to this metatable */
  luaL_register(L, AGENA_DYVECLIBNAME, dyveclib);
  return 1;
}

