/*
** $Id: fastmath.c,v 1.67 2005/08/26 17:36:32 roberto Exp $
** Approximations
** See Copyright Notice in agena.h
*/


#include <stdlib.h>
#include <math.h>
#include <stdint.h>  /* for UINT32_MAX */

#define fastmath_c
#define LUA_LIB

#include "agena.h"

#include "agnxlib.h"
#include "agenalib.h"
#include "agnconf.h"
#include "agncmpt.h"  /* for trunc and isfinite, and FLT_EVAL_METHOD constant, and off64* types */

/* For unknown reasons I am too tired to clarify, we import the *GET_*, *SET_*, *INSERT*, *EXTRACT*,
   etc. Sun Microsystems macros from the following header file instead of agnhlps.h: */
#include "sunpro.h"

#define AGENA_LIBVERSION	"fastmath 1.1.1 for Agena as of May 13, 2019\n"

#if !(defined(LUA_DOS) || defined(__OS2__) || defined(LUA_ANSI))
#define AGENA_FASTMATHLIBNAME "fastmath"
LUALIB_API int (luaopen_fastmath) (lua_State *L);
#endif


/* Approximates the inverse root 1/root(x, degree) using the Quake III method. x is the radicant, degree the
   degree-th root which by default is 2. n is the number of iterations to be conducted and by default is 2^degree.
   xhalf is the internal equivalent of x, 0.5*x by default. The greater the degree, the less accurate is the result.

   See: https://stackoverflow.com/questions/11644441/fast-inverse-square-root-on-x64,
   https://www.xaymar.com/2016/12/17/what-is-the-fastest-way-to-get-an-inverse-square-root-part-2
   Modified. 2.10.4
   Must use a union instead of *(<type> *)&y, and the proposed int64_t type given in the web pages
   listed above is wrong, anyway. */
static int fastmath_invroot (lua_State *L) {
  lua_Number x, y, xhalf;
  double_cast d;
  int i, n, degree;
  x = agn_checknumber(L, 1);
  if (x ==0) y = 0;
  else if (x < 0) y = AGN_NAN;
  else {
    degree = agnL_optposint(L, 2, 2);
    n = agnL_optposint(L, 3, tools_intpow(2, degree));
    xhalf = agnL_optnumber(L, 4, 0.5*x);  /* internal equivalent for x, 0.5*x by default */
    /* int64_t i = *(int64_t *)&y; does not work with GCC 6.3.0, especially not with int64_t's */
    d.f = x;
    /* The magic number 0x5fe6eb50c7b537a9 is for doubles is from https://cs.uwaterloo.ca/~m32rober/rsqrt.pdf */
    /* d.i = 0x5fe6eb50c7b537a9 - (d.i >> 1);  */ /* first guess of 1/sqrt(x) */
    d.i = agnL_optnumber(L, 5, 0x5fe6eb50c7b537a9ll) - (d.i >> 1);  /* 0x5fe6eb50c7b537a9ll taken from:
      https://www.xaymar.com/2016/12/17/what-is-the-fastest-way-to-get-an-inverse-square-root-part-2/ */
    /* y = *(lua_Number *)&i;  does not work with GCC 6.3.0  */
    y = d.f;
    for (i=0; i < n; i++)  /* 4 Newton iterations to approximate 1/sqrt(x) */
      y = y*(1.5 - (xhalf*tools_intpow(y, degree)));  /* 1.5 = `threehalfs` */
    /* to compute the square root, now execute: y = ((1/y) + (x*y))*0.5; (1 additional Heron iteration has shown
       to maximise the precision), see: http://www.reactos.org/pipermail/ros-diffs/2015-May/057628.html */
  }
  lua_pushnumber(L, y);  /* square y to return 1/x */
  return 1;
}


/* Approximates 1/sqrt(x) using Quakes Fast Inverse Square Root method. Five percent faster than the inverse of
   GCC's builtin sqrt. In the range -1000 .. 1000 with step size 0.1, mean error is 5.96994e-5, median error is 5.008699e-5.

   See: https://betterexplained.com/articles/understanding-quakes-fast-inverse-square-root */
static int fastmath_invsqrt (lua_State *L) {  /* new in 2.14.6 */
  float x, xhalf;
  ieee_float_shape_signedint u;
  x = agn_checknumber(L, 1);
  xhalf = 0.5f * x;
  u.f = x;                        /* store floating-point bits in integer */
  u.i = 0x5f3759df - (u.i >> 1);  /* initial guess for Newton's method */
  x = u.f;                        /* convert new bits into float */
  x = x*(1.5f - xhalf*x*x);       /* One round of Newton's method */
  lua_pushnumber(L, x);
  return 1;
}


/* Approximates of the reciprocal of its argument of type number. The return is a number. The function is
   purely experimental.
   See: https://stackoverflow.com/questions/9939322/fast-1-x-division-reciprocal/39714493 */
static int fastmath_reciprocal (lua_State *L) {
  double_cast d;
  d.f = agn_checknumber(L, 1);
  d.i = (0xbfcdd6a18f6a6f52ULL - d.i) >> 1;
  lua_pushnumber(L, d.f*d.f);
  return 1;
}


/* Approximates of the square root using two numeric methods. The return is a number. The function is
   purely experimental.
   See: http://bits.stephan-brumme.com/squareRoot.html */
static int fastmath_sqroot (lua_State *L) {
  ieee_float_shape_type d;
  double_cast e;
  double x = agn_checknumber(L, 1);
  if (x < 0) {  /* 2.12.0 RC 3 change */
    lua_pushundefined(L);
    return 1;
  }
  d.value = x;
  e.f = x;
  /* adjust bias */
  d.word += (127 << 23);
  /* approximation of square root */
  d.word >>= 1;
  /* see: http://www.reactos.org/pipermail/ros-diffs/2015-May/057628.html */
  e.i = 0x5fe6eb50c7b537a9ll - (e.i >> 1);
  lua_pushnumber(L, 1/e.f);
  lua_pushnumber(L, d.value);
  return 2;
}


/*
http://jsperf.com/mysin-vs-sin/9
https://www.desmos.com/calculator/8nkxlrmp7a
http://devmaster.net/posts

See: https://web.archive.org/web/20161223122133/http://http.developer.nvidia.com:80/Cg/asin.html
*/
static int fastmath_sincosfast (lua_State *L) {  /* 2.11.0, 15 percent faster than calling Agena's sin and cos operators */
  float x, s, c;
  x = agn_checknumber(L, 1);
  if (lua_gettop(L) == 1)
    tools_sincosfast(x, &s, &c);
  else {
    /* taken from http://lab.polygonal.de/2007/07/18/fast-and-accurate-sinecosine-approximation/ */
    /* always wrap input angle to -PI .. PI */
    x = tools_reducerange(x, -PI, PI);
    /* compute sine */
    s = INVPIO4 * x - (x)*INVPISQO4*x*x;
    s += 0.225 * (luai_numsignum(s)*s*s - s);  /* 2.15.0 change */
    /* compute cosine: sin(x + PI/2) = cos(x) */
    x += PIO2;
    if (x > PI) x -= PI2;
    c = INVPIO4 * x - luai_numsignum(x)*INVPISQO4*x*x;  /* 2.15.0 change */
    c += 0.225 * (luai_numsignum(c)*c*c - c);  /* 2.15.0 change */
  }
  lua_pushnumber(L, s);
  lua_pushnumber(L, c);
  return 2;
}


/* Returns an approximation of the sine of the given number x, in radians. Depending on the CPU used, it is up to 40 percent faster
  than calling the `sin` operator. 2.14.6 */

static int fastmath_sinfast (lua_State *L) {
  lua_pushnumber(L, tools_sinfast(agn_checknumber(L, 1)));
  return 1;
}


/* Returns an approximation of the cosine of the given number x, in radians. Depending on the CPU used, it is up to 40 percent faster
  than calling the `cos` operator. 2.14.6 */

static int fastmath_cosfast (lua_State *L) {
  lua_pushnumber(L, tools_cosfast(agn_checknumber(L, 1)));
  return 1;
}


/* Returns an approximation of the tangent of the given number x, in radians. Depending on the CPU used, it is up to 40 percent faster
   than calling the `tan` operator. 2.14.6 */

static int fastmath_tanfast (lua_State *L) {
  lua_Number x = agn_checknumber(L, 1);
  lua_pushnumber(L, tools_sinfast(x)/tools_cosfast(x));
  return 1;
}


/* Approximates sqrt(x) but is much slower than the sqrt operator. If x < 0, returns `undefined`.
   In the range 0 .. 1000 with step size 0.1, median error is 0.00791 and mean error is 0.00829. */
static int fastmath_sqrtfast (lua_State *L) {  /* 2.11.0 */
  lua_pushnumber(L, tools_sqrtfast(agn_checknumber(L, 1)));
  return 1;
}


static int fastmath_sqrtapx (lua_State *L) {  /* 2.15.1 */
  lua_pushnumber(L, tools_sqrtapx(agn_checknumber(L, 1)));
  return 1;
}


/* The function approximates log2(x) for number x, and returns a number. It is around 18 percent faster than GCC's builtin log2
   but less acuurate. If x <= 0, the result is wrong.

   In the range 0 .. 1000 with step size 0.1, median error is 0.0000824684 and mean error is 0.00007639151. */
static int fastmath_lbfast (lua_State *L) {  /* new in 2.14.6 */
  lua_pushnumber(L, tools_lbfast(agn_checknumber(L, 1)));
  return 1;
}


/* The function approximates ln(x) for number x, and returns a number. It is around 18 percent faster than GCC's builtin log2
   but less acuurate. If x <= 0, the result is wrong. 18 percent faster than sun_log.

   In the range 0 .. 1000 with step size 0.1, median error is 0.00005715 and mean error is 0.00005294. */
static int fastmath_lnfast (lua_State *L) {  /* 2.14.6 */
  lua_pushnumber(L, (float)LN2 * tools_lbfast(agn_checknumber(L, 1)));
  return 1;
}


/* Returns the hypotenuse of the two numbers x and y; the return is a number. The function is sixty percent faster than `hypot`,
  but prone two round-off errors. 60 percent faster than `hypot', 27 percent faster than the Agena implementation sqrt(x**2 + y**2). */
static int fastmath_hypotfast (lua_State *L) {  /* 2.14.4 */
  lua_Number x, y;
  x = agn_checknumber(L, 1);
  y = agn_checknumber(L, 2);
  lua_pushnumber(L, sqrt(x*x + y*y));
  return 1;
}


static int fastmath_pow2fast (lua_State *L) {  /* 2.14.6 */
  lua_pushnumber(L, tools_pow2fast(agn_checknumber(L, 1)));
  return 1;
}


/* Approximates e^x with enough precission (mean error = 7.5476762e-6) for -infinity < x < 1.
   For 1 < x < 10, mean error = 0.05514. For x > 10 accuracy becomes quite bad. */
static int fastmath_expfast (lua_State *L) {  /* 2.14.6 */
  lua_pushnumber(L, tools_pow2fast((float)agn_checknumber(L, 1) * (float)INVLN2));
  return 1;
}


/* For reference tests only:

static int fastmath_sunsin (lua_State *L) {
  lua_pushnumber(L, sun_sin(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccsin (lua_State *L) {
  lua_pushnumber(L, sin(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_suncos (lua_State *L) {
  lua_pushnumber(L, sun_cos(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gcccos (lua_State *L) {
  lua_pushnumber(L, cos(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_suntan (lua_State *L) {
  lua_pushnumber(L, sun_tan(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gcctan (lua_State *L) {
  lua_pushnumber(L, tan(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_sunlog (lua_State *L) {
  lua_pushnumber(L, sun_log(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gcclog (lua_State *L) {
  lua_pushnumber(L, log(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_sunexp (lua_State *L) {
  lua_pushnumber(L, sun_exp(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccexp (lua_State *L) {
  lua_pushnumber(L, exp(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_sunlog2 (lua_State *L) {
  lua_pushnumber(L, sun_log2(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gcclog2 (lua_State *L) {
  lua_pushnumber(L, log2(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccsqrt (lua_State *L) {
  lua_pushnumber(L, sqrt(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_sunasinh (lua_State *L) {
  lua_pushnumber(L, sun_asinh(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccasinh (lua_State *L) {
  lua_pushnumber(L, asinh(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccacosh (lua_State *L) {
  lua_pushnumber(L, acosh(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccasin (lua_State *L) {
  lua_pushnumber(L, asin(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccacos (lua_State *L) {
  lua_pushnumber(L, acos(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_gccatan (lua_State *L) {
  lua_pushnumber(L, atan(agn_checknumber(L, 1)));
  return 1;
}

static int fastmath_sunpow (lua_State *L) {
  lua_pushnumber(L, sun_pow(agn_checknumber(L, 1), agn_checknumber(L, 2)));
  return 1;
}

static int fastmath_gccatanh (lua_State *L) {
  lua_pushnumber(L, atanh(agn_checknumber(L, 1)));
  return 1;
} */


static const luaL_Reg fastmathlib[] = {
  {"cosfast", fastmath_cosfast},              /* added on January 18, 2019 */
  {"expfast", fastmath_expfast},              /* added on January 21, 2019 */
  {"hypotfast", fastmath_hypotfast},          /* added on January 01, 2019 */
  {"invroot", fastmath_invroot},              /* added on August 01, 2017 */
  {"invsqrt", fastmath_invsqrt},              /* added on January 21, 2019 */
  {"lbfast", fastmath_lbfast},                /* added on June 16, 2018 */
  {"lnfast", fastmath_lnfast},                /* added on January 21, 2019 */
  {"pow2fast", fastmath_pow2fast},            /* added on January 22, 2019 */
  {"reciprocal", fastmath_reciprocal},        /* added on August 01, 2017 */
  {"sincosfast", fastmath_sincosfast},        /* added on October 17, 2017 */
  {"sinfast", fastmath_sinfast},              /* added on January 18, 2019 */
  {"sqroot", fastmath_sqroot},                /* added on August 07, 2017 */
  {"sqrtfast", fastmath_sqrtfast},            /* added on October 17, 2017 */
  {"sqrtapx", fastmath_sqrtapx},              /* added on May 23, 2019 */
  {"tanfast", fastmath_tanfast},              /* added on January 18, 2019 */
  /* reference functions
  {"gccacos", fastmath_gccacos},
  {"gccacosh", fastmath_gccacosh},
  {"gccasin", fastmath_gccasin},
  {"gccasinh", fastmath_gccasinh},
  {"gccatanh", fastmath_gccatanh},
  {"gccatan", fastmath_gccatan},
  {"gcccos", fastmath_gcccos},
  {"gccexp", fastmath_gccexp},
  {"gcclog", fastmath_gcclog},
  {"gcclog2", fastmath_gcclog2},
  {"gccsin", fastmath_gccsin},
  {"gccsqrt", fastmath_gccsqrt},
  {"gcctan", fastmath_gcctan},
  {"sunasinh", fastmath_sunasinh},
  {"suncos", fastmath_suncos},
  {"sunexp", fastmath_sunexp},
  {"sunlog", fastmath_sunlog},
  {"sunlog2", fastmath_sunlog2},
  {"sunpow", fastmath_sunpow},
  {"sunsin", fastmath_sunsin},
  {"suntan", fastmath_suntan}, */
  {NULL, NULL}
};


/*
** Open fastmath library
*/
LUALIB_API int luaopen_fastmath (lua_State *L) {
  luaL_register(L, AGENA_FASTMATHLIBNAME, fastmathlib);
  return 1;
}

