/* Dynamic Vector Array, initiated September 27, 2017

   Based on: http://eddmann.com/posts/implementing-a-dynamic-vector-array-in-c, modified

   In GCC, when linking to this library, use
   -Wno-int-to-pointer-cast -Wno-pointer-to-int-cast switches. With type off64_t, this does not causs memory problems

   You CANNOT pass DOUBLES or other simple 8-byte data types via (void *) arguments, since pointers are usually 4 bytes wide.
   Put them into unions or arrays, alternatively. Casting doubles to pointers like "double *p = (double *)num" and back
   to "double num = *p;" is not portable.

   See also: https://stackoverflow.com/questions/6575340/converting-double-to-void-in-c
             https://stackoverflow.com/questions/12700384/convert-void-to-double */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>  /* for int32_t */

#define vector_c

#include "vector.h"
#include "agnhlps.h"

/* The growth pattern is:  0, 4, 8, 16, 25, 35, 46, 58, 72, 88, etc.
   Taken from: https://news.ycombinator.com/item?id=1996857 */
/* #define newsize(x) (x + (x >> 3) + (x < 9 ? 3 : 6)) */
#define newsize(x)    (x * 2)
#define ceillog2(x)   (int)(sun_ilogb(x) + 1)

int vector_init (vector *v, size_t itemsize) {
  v->capacity = VECTOR_INICAPA;
  v->size = 0;
  v->itemsize = itemsize;  /* sizeof(type) */
  v->data = malloc(itemsize * v->capacity);
  return v->data ? 0 : 1;  /* 0 = success, 1 = failure */
}


int vector_size (vector *v) {
  return v->size;
}


static int vector_resize (vector *v, size_t capacity) {
  /* void **items = realloc(v->items, sizeof(void *) * capacity); */
  void **data = realloc(v->data, v->itemsize * capacity);
  if (data) {
    v->data = data;
    v->capacity = capacity;
  }
  return data ? 0 : 1;  /* 0 = success, 1 = failure */
}


/* Append ! call this function as follows: vector_add(&v, (void *)z); where v is the vector, and z the value */
int vector_append (vector *v, void *item) {
  if (v->capacity == v->size) {
    if (vector_resize(v, newsize(v->capacity))) return 1;
  }
  v->data[v->size++] = item;
  return 0;  /* 0 = success, 1 = failure */
}


void vector_set (vector *v, int index, void *item) {
  if (index >= 0 && index < v->size)
    v->data[index] = item;
}


/* call like this: (off64_t)vector_get(&v, index) */
void *vector_get (vector *v, int index) {
  if (index >= 0 && index < v->size)
    return v->data[index];
  return NULL;
}


int vector_delete (vector *v, int index) {
  int i;
  if (index < 0 || index >= v->size) return 2;  /* out of range */
  v->data[index] = NULL;
  for (i=index; i < v->size - 1; i++) {
    v->data[i] = v->data[i + 1];
    v->data[i + 1] = NULL;
  }
  v->size--;
  /* if (v->size > 0 && v->size == v->capacity/4) { */
  if (v->size > 0) {  /* changes 2.15.4 */
    size_t p = 2 << ( ( (v->size < VECTOR_INICAPA) ? ceillog2(VECTOR_INICAPA - 1) : ceillog2(v->size) ) - 1 );
printf("v->size=%d p=%d\n", v->size, p);
    if (vector_resize(v, p)) return 1;
  }
  return 0;  /* 0 = success, 1 = failure */
}


/* can also insert to the tail */
int vector_insert (vector *v, int index, void *item) {  /* 2.15.4, written by aw */
  int i;
  if (index < 0 || index > v->size) return 2;  /* out of range */
  if (index == v->size) return vector_append(v, item);  /* append */
  if (v->capacity == v->size) {
    if (vector_resize(v, newsize(v->capacity))) return 1;
  }
  for (i=v->size; i > index; i--) {
    v->data[i] = v->data[i - 1];
  }
  v->data[index] = item;
  v->size++;
  return 0;  /* 0 = success, 1 = failure */
}


void vector_free (vector *v) {
  free(v->data);
  v->data = NULL;
}

